import os
import pwd
from ConfigParser import RawConfigParser, NoOptionError


class Config():

    """
    Load and access the carbonate configuration.
    """

    def __init__(self, config_file):
        self.config_file = config_file
        self.config = RawConfigParser()
        self.config.read(config_file)

    def clusters(self):
        """Return the clusters defined in the config file."""
        return self.config.sections()

    def destinations(self, cluster='main'):
        """Return a list of destinations for a cluster."""
        if not self.config.has_section(cluster):
            raise SystemExit("Cluster '%s' not defined in %s"
                             % (cluster, self.config_file))
        destinations = self.config.get(cluster, 'destinations')
        return destinations.replace(' ', '').split(',')

    def replication_factor(self, cluster='main'):
        """Return the replication factor for a cluster as an integer."""
        if not self.config.has_section(cluster):
            raise SystemExit("Cluster '%s' not defined in %s"
                             % (cluster, self.config_file))
        return int(self.config.get(cluster, 'replication_factor'))

    def ssh_user(self, cluster='main'):
        """Return the ssh user for a cluster or current user if undefined."""
        if not self.config.has_section(cluster):
            raise SystemExit("Cluster '%s' not defined in %s"
                             % (cluster, self.config_file))
        try:
            return self.config.get(cluster, 'ssh_user')
        except NoOptionError:
            return pwd.getpwuid(os.getuid()).pw_name

    def whisper_lock_writes(self, cluster='main'):
        """Lock whisper files during carbon-sync."""
        if not self.config.has_section(cluster):
            raise SystemExit("Cluster '%s' not defined in %s"
                             % (cluster, self.config_file))
        try:
            return bool(self.config.get(cluster, 'whisper_lock_writes'))
        except NoOptionError:
            return False



import os
import sys
import logging
import shutil
import subprocess
from time import time
from datetime import timedelta
from tempfile import mkdtemp, NamedTemporaryFile
from shutil import rmtree
from whisper import CorruptWhisperFile

from .fill import fill_archives


def sync_from_remote(sync_file, remote, staging, rsync_options):
    try:
        try:
            os.makedirs(os.path.dirname(staging))
        except OSError:
            pass

        cmd = " ".join(['rsync', rsync_options, '--files-from',
                        sync_file.name, remote, staging
                        ])

        print "  - Rsyncing metrics"

        proc = subprocess.Popen(cmd,
                                shell=True,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT)

        for l in iter(proc.stdout.readline, b''):
            sys.stdout.write(l)
            sys.stdout.flush()
    except subprocess.CalledProcessError as e:
        logging.warn("Failed to sync from %s! %s" % (remote, e))


def sync_batch(metrics_to_heal, lock_writes=False):
    batch_start = time()
    sync_count = 0
    sync_total = len(metrics_to_heal)
    sync_avg = 0.1
    sync_elapsed = 0
    sync_remain = 'n/a'

    for (staging, local) in metrics_to_heal:
        sync_count += 1
        sync_start = time()
        sync_percent = float(sync_count) / sync_total * 100
        status_line = "  - Syncing %d of %d metrics. " \
                      "Avg: %fs  Time Left: %ss (%d%%)" \
                      % (sync_count, sync_total, sync_avg,
                         sync_remain, sync_percent)
        print status_line

        # Do not try healing data past the point they were rsync'd
        # as we would not have new points in staging anyway.
        heal_metric(staging, local, end_time=batch_start,
                    lock_writes=lock_writes)

        sync_elapsed += time() - sync_start
        sync_avg = sync_elapsed / sync_count
        sync_remain_s = sync_avg * (sync_total - sync_count)
        sync_remain = str(timedelta(seconds=sync_remain_s))

    batch_elapsed = time() - batch_start
    return batch_elapsed


def heal_metric(source, dest, start_time=0, end_time=None, overwrite=False,
                lock_writes=False):
    if end_time is None:
        end_time = time()
    try:
        with open(dest):
            try:
                # fill_archives' start and end are the opposite
                # of what you'd expect
                fill_archives(
                    source, dest, startFrom=end_time, endAt=start_time,
                    overwrite=overwrite, lock_writes=lock_writes)
            except CorruptWhisperFile as e:
                if e.path == source:
                    # The source file is corrupt, we bail
                    logging.warn("Source file corrupt, skipping: %s" % source)
                else:
                    # Do it the old fashioned way...possible data loss
                    logging.warn("Overwriting corrupt file: %s" % dest)
                    try:
                        os.makedirs(os.path.dirname(dest))
                    except os.error:
                        pass
                    try:
                        # Make a backup of corrupt file
                        corrupt = dest + ".corrupt"
                        shutil.copyfile(dest, corrupt)
                        logging.warn("Corrupt file saved as %s" % corrupt)
                        shutil.copyfile(source, dest)
                    except IOError as e:
                        logging.warn("Failed to copy %s! %s" % (dest, e))
            except Exception as e:
                logging.warn("Exception during heal: %s" % str(e))
                logging.warn("Skipping heal: %s => %s" % (source, dest))
    except IOError:
        try:
            os.makedirs(os.path.dirname(dest))
        except os.error:
            pass
        try:
            shutil.copyfile(source, dest)
        except IOError as e:
            logging.warn("Failed to copy %s! %s" % (dest, e))


def run_batch(metrics_to_sync, remote, local_storage, rsync_options,
              remote_ip, dirty, lock_writes=False):
    staging_dir = mkdtemp(prefix=remote_ip)
    sync_file = NamedTemporaryFile(delete=False)

    metrics_to_heal = []

    staging = "%s/" % (staging_dir)

    for metric in metrics_to_sync:
        staging_file = "%s/%s" % (staging_dir, metric)
        local_file = "%s/%s" % (local_storage, metric)
        metrics_to_heal.append((staging_file, local_file))

    sync_file.write("\n".join(metrics_to_sync))
    sync_file.flush()

    rsync_start = time()

    sync_from_remote(sync_file, remote, staging, rsync_options)

    rsync_elapsed = (time() - rsync_start)

    merge_elapsed = sync_batch(metrics_to_heal, lock_writes=lock_writes)

    total_time = rsync_elapsed + merge_elapsed

    print "    --------------------------------------"
    print "    Rsync time: %ss" % rsync_elapsed
    print "    Merge time: %ss" % merge_elapsed
    print "    Total time: %ss" % total_time

    # Cleanup
    if dirty:
        print "    dirty mode: left temporary directory %s" % staging_dir
    else:
        rmtree(staging_dir)

    os.unlink(sync_file.name)



import logging
import os
import whisper

AGGREGATION = {
    'last': 'last',
    'g': 'last',

    'sum': 'sum',
    'c': 'sum',
    'C': 'sum',

    'average': None,  # set by graphite
    'ms': None,       # set by graphite
    'h': None,        # set by graphite
    'internal': None
}


def setAggregation(path, mode):

    if not mode:
        return 0

    if not os.path.exists(path):
        return 0

    try:
        whisper.setAggregationMethod(path, mode)
        return 1
    except whisper.WhisperException, exc:
        logging.warning("%s failed (%s)" % (path, str(exc)))



import os
import re


def listMetrics(storage_dir, follow_sym_links=False, metric_suffix='wsp'):
    metric_regex = re.compile(".*\.%s$" % metric_suffix)

    storage_dir = storage_dir.rstrip(os.sep)

    for root, dirnames, filenames in os.walk(storage_dir,
                                             followlinks=follow_sym_links):
        for filename in filenames:
            if metric_regex.match(filename):
                root_path = root[len(storage_dir) + 1:]
                m_path = os.path.join(root_path, filename)
                m_name, m_ext = os.path.splitext(m_path)
                m_name = m_name.replace('/', '.')
                yield m_name
